/*
   ESP8266_WeatherStation_menu.cpp
   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/
// For Adafruit HUZZAH ESP8266
/**
   This sketch accepts serial inputs on Serial consistion of BLE Senor names and values
   Looks for WL_1, and W_1, sensors.
   WL_1 is the outside sensor, W_1 is the inside sensor
   WL_1 has a lux level, for a future project, in addition to temp,rh,pressure

  Sensor advertised names start with
  T_.. for Temp(C) sensors, H_.. for Temp(C),Humditiy(%) sensors, W_.. for Temp(C),Humdity(%),Barometer(hPa) sensors, WL_.. for Temp(C),Humdity(%),Barometer(hPa),Lux level(Lux)
  keep this LOCAL_NAME short as it uses up available advertising data space (see below)
  Format of advertized name
  For T_.. devices   LOCAL_NAME,degC
  For H_.. devices   LOCAL_NAME,degC,%RH
  For W_.. devices   LOCAL_NAME,degC,%RH,hPa
  For WL_.. devices  LOCAL_NAME,degC,%RH,hPa,Lux
*/
// for ESP8266 FS 2M OTA ~1019K (default)
// Needs Huge APP, 2M APP, 1M SPIFF  Partition setting

#if defined (ESP32)
#include <WiFi.h>
#elif defined (ESP8266)
#include <ESP8266WiFi.h>
#else
#error "Only ESP32 and ESP8266 supported"
#endif


#include <pfodEEPROM.h>
#include "display.h"
#include "ntpSupport.h"

//#define DEBUG
static Stream *debugPtr = NULL;
// ========== settings for the Auto WiFi connect ====================================
// Adafruit ESP8266 HUZZAH
int ledPin = 0; //for Adafruit ESP8266 HUZZAH - onboard Led connected to GPIO0
bool highForLedOn = false; // need to make output low to turn Adafruit ESP8266 HUZZAH's and ESP-01s and ESP-01 Led ON
// allow space for pfodSecurity EEPROM usage
size_t eepromOffset = 40; // if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data
// ==============================================================


void start_pfodServer();
void handle_pfodServerConnection();

#include "ESPAutoWiFiConfig.h"  // download ESPAutoWiFiConfig.zip from  https://www.forward.com.au/pfod/ESPAutoWiFiConfig/index.html
#include "pfodLinkedPointerList.h" // iterable linked list of pointers to objects
#include "LastSeen.h"  // class for storing on linked list
#include "SafeString.h"

#if defined (ESP32)
// defined in FreeRTOS
SemaphoreHandle_t listOfLastSeen_Semaphore = NULL; // used to copy from listOf_BLE_LastSeen to listOfLastSeen
#elif defined (ESP8266)
#else
#endif

bool listUpdated = false; // set to true when copy from listOf_BLE_LastSeen to listOfLastSeen done, cleared when listOfLastSeen processed  // protected by listOfLastSeen_Semaphore

// only BLE thread writes to these structs,
// loop() only reads from these and copies data for updates
// if BLE has lock on struct, loop() skips read and uses previous data
pfodLinkedPointerList<LastSeen> listOf_BLE_LastSeen; // updated by BLE
pfodLinkedPointerList<LastSeen> listOfLastSeen; // filled from BLE after each scan
const unsigned long BAUD_RATE = 115200ul;

// returns NULL if not found OR if list Ptr is NULL OR name is NULL
LastSeen* getLastSeen(const char* name, pfodLinkedPointerList<LastSeen>* listPtr) {
  if (!listPtr) {
    return NULL;
  }
  if (!name) {
    return NULL;
  }
  cSF(sfName, strlen(name));
  sfName = name;
  return getLastSeen(sfName, listPtr);
}

// returns NULL if not found OR if list Ptr is NULL
LastSeen* getLastSeen(SafeString &name, pfodLinkedPointerList<LastSeen>* list) {
  if (!list) {
    return NULL;
  }
  LastSeen *devicePtr = list->getFirst();
  while (devicePtr) {
    if (name.startsWith(devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = list->getNext();
  }
  return devicePtr; // null not found
}



int devicesFoundThisScan = 0;

// this call is protected by listOfLastSeen_Semaphore (if non-null)
void copyListOfLastSeen() { // sets listUpdated if updated
  // update listOfLastSeen for WeatherStation
  // use semaphore here, this is certainly not an ISR
  LastSeen *BLE_devicePtr = listOf_BLE_LastSeen.getFirst();
  while (BLE_devicePtr != NULL) {
    LastSeen *devicePtr = getLastSeen(BLE_devicePtr->getDeviceName(), &listOfLastSeen);
    if (!devicePtr) {
      // add it
      devicePtr = new LastSeen(BLE_devicePtr->getDeviceName());
      listOfLastSeen.add(devicePtr); // addit to WeatherStation list
    }
    // copy from BLE to listOfLastSeen if different
    if (devicePtr->getLastSeen() != BLE_devicePtr->getLastSeen()) {
      devicePtr->updateLastSeen(BLE_devicePtr->getLastSeen());
      devicePtr->setAdvertisedName(BLE_devicePtr->getAdvertisedName());
      listUpdated = true;
    }
    BLE_devicePtr = listOf_BLE_LastSeen.getNext();
  }
  devicesFoundThisScan = 0; // have copied the devices found, interested devices only
}
void setUpWiFiServices() {
  start_pfodServer();
}
void setup() {
  EEPROM.begin(512);  // only use 20bytes for pfodSecurity but reserve 512 (pfodWifiConfig uses more)
  Serial.begin(115200);

#ifdef DEBUG
  Serial.println();
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  SafeString::setOutput(Serial); // for debug error msgs
  debugPtr = &Serial;
  debugPtr->println();
#endif
  /* Initialise wifi module */
#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, eepromOffset)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

  setDisplayDebug(debugPtr); // debugPtr can be NULL OK
  initializeDisplay(); // for pfodApp
  setNtpSupportDebug(debugPtr);
  initializeNtpSupport();
  
#if defined (ESP32)
  /* Create a mutex type semaphore. */
  listOfLastSeen_Semaphore = xSemaphoreCreateMutex();
  if (!listOfLastSeen_Semaphore) {
    if (debugPtr) {
      debugPtr->println("xSemaphoreCreateMutex failed, just copy without mutex protection");
    }
  }
#elif defined (ESP8266)
  // listOfLastSeen_Semaphore not used
#endif
  // add the names we are interested in NOTE the trailing , so that T_1, does not match T_11,
  LastSeen *devicePtr = new LastSeen(getOutsideDevice()); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
  listOf_BLE_LastSeen.add(devicePtr);
  // add other devices here that you want to scan for
  devicePtr = new LastSeen(getInsideDevice());
  listOf_BLE_LastSeen.add(devicePtr);

  if (debugPtr) {
    debugPtr->println(" ================ ");
    debugPtr->print("Scanning for ");
    LastSeen *lastSeenPtr = listOf_BLE_LastSeen.getFirst();
    if (lastSeenPtr == NULL) {
      debugPtr->println(" Error: no devices specified.");
    } else {
      while (lastSeenPtr != NULL) {
        debugPtr->print(lastSeenPtr->getDeviceName()); debugPtr->print(" ");
        lastSeenPtr = listOf_BLE_LastSeen.getNext();
      }
    }
    debugPtr->println();
    debugPtr->println(" ================ ");
  }

  setUpWiFiServices(); // do this first. Get continual reboots if create BLE task first and then call this
}

const static size_t MAX_BLE_NAME_SIZE = 32;

void processSensorData(SafeString &sensorData) {
  // have a new line or sensorData filled up
  devicesFoundThisScan = 0;
  sensorData.trim();
  if (sensorData.length() > MAX_BLE_NAME_SIZE) {
    // not advertising data (debug data?) just echo
#ifdef DEBUG
    Serial.print("Input not sensor data -- "); Serial.println(sensorData);
#endif
    return;
  } //else {
  // search listOfLastSeen
  LastSeen *devicePtr = getLastSeen(sensorData, &listOf_BLE_LastSeen);
  if (!devicePtr) {
    // not found
#ifdef DEBUG_BLE_SCANNING
    if (debugPtr) {
      debugPtr->print("Not scanning for "); debugPtr->println(sfName);
    }
#endif
    return;
  }
  // else found it check if seen recently
  unsigned long now = millis();
  devicePtr->updateLastSeen(now);
  devicePtr->setAdvertisedName(sensorData.c_str()); // save the full advertised data
  if (debugPtr) {
    debugPtr->print("Scan found "); debugPtr->println(sensorData);
  }
  devicesFoundThisScan++;
  sensorData.clear(); // for next read
}

void processSensorResults() {
  if (devicesFoundThisScan) {
#ifdef DEBUG_BLE_SCANNING
    if (debugPtr) {
      debugPtr->print("Devices found: ");
      debugPtr->println(devicesFoundThisScan);
    }
#endif
#if defined (ESP32)
    if (!listOfLastSeen_Semaphore) {
      // just copy
      copyListOfLastSeen(); // clears devicesFoundThisScan
    } else {// take it wait a little
      if ( xSemaphoreTake( listOfLastSeen_Semaphore, ( TickType_t ) 10 ) == pdTRUE ) {
        copyListOfLastSeen(); // clears devicesFoundThisScan
        xSemaphoreGive( listOfLastSeen_Semaphore );
      } // else devicesFoundThisScan not cleared will try and copy next time around
    }
#elif defined (ESP8266)
    copyListOfLastSeen(); // clears devicesFoundThisScan
#else
#endif
  }

#if defined (ESP32)
  if (!listOfLastSeen_Semaphore) {
    // just copy
    if (listUpdated) {
      listUpdated = false;
      processListOfLastSeen(listOfLastSeen); //
    }

  } else {// take it wait a little
    if ( xSemaphoreTake( listOfLastSeen_Semaphore, ( TickType_t ) 10 ) == pdTRUE ) {
      if (listUpdated) {
        listUpdated = false;
        processListOfLastSeen(listOfLastSeen);
      }
      xSemaphoreGive(listOfLastSeen_Semaphore );
    } // else devicesFoundThisScan not cleared will try and copy next time around
  }
#elif defined (ESP8266)
  if (listUpdated) {
    listUpdated = false;
    processListOfLastSeen(listOfLastSeen);
  }
#else
#endif
}

cSF(sfSensorDataIn, 50); // max 32
void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }
  // see if BLE board has sent any data
  if (sfSensorDataIn.readUntil(Serial, '\n')) {
    processSensorData(sfSensorDataIn);
    // after all the copies
    processSensorResults();
  }
  processDisplay(); // needs to be called each loop
  processNTP();
  handle_pfodServerConnection();
}
// ------------------ trival telnet server -------------------------
//how many clients should be able to telnet to this ESP8266
#define MAX_SRV_CLIENTS 1

WiFiServer pfodServer(4989);
WiFiClient pfodServerClients[MAX_SRV_CLIENTS];

void start_pfodServer() {
  pfodServer.begin();
  pfodServer.setNoDelay(true);
  if (debugPtr) {
    debugPtr->print("pfodServer Ready to Use on ");
    debugPtr->print(WiFi.localIP());
    debugPtr->println(":4989");
  }
}

void handle_pfodServerConnection() {
  //check if there are any new clients
  uint8_t i;
  if (pfodServer.hasClient()) {
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      //find free/disconnected spot
      if (!pfodServerClients[i] || !pfodServerClients[i].connected()) {
        if (pfodServerClients[i]) {
          pfodServerClients[i].stop();
        }
        pfodServerClients[i] = pfodServer.available();
        if (!pfodServerClients[i]) {
          if (debugPtr) {
            debugPtr->println("available broken");
          }
        } else {
          if (debugPtr) {
            debugPtr->print("New client: ");
            debugPtr->print(i); debugPtr->print(' ');
            debugPtr->println(pfodServerClients[i].remoteIP());
          }
          connect_pfodParser(&pfodServerClients[i]);
          break;
        }
      }
    }
    if (i >= MAX_SRV_CLIENTS) {
      //no free/disconnected spot so reject
      pfodServer.available().stop();
    }
  }
  //check clients for data
  for (i = 0; i < MAX_SRV_CLIENTS; i++) {
    if (pfodServerClients[i] && pfodServerClients[i].connected()) {
      handle_pfodParser();
    }
    else {
      if (pfodServerClients[i]) {
        pfodServerClients[i].stop();
      }
    }
  }
}
